/* WIN32.C - Win16 versions of Win32 APIs for Visual Basic.
 * The Win16 routines are C-callable. Some APIs return
 * pointers or other data that Basic can't handle directly.
 * For these there are Basic wrappers that pass or return
 * return data in a Basic-friendly way.

		Win32 Routine			Basic-Friendly Version
		-------------			----------------------
		FlushFileBuffers		None (see Basic version)
		GetDiskFreeSpace		None
		GetDriveType			None
		GetFullPathName 		GetFullPath
		GetLastError			None
		GetLogicalDrives		None
		GetLogicalDriveStrings	None
		GetTempFileName 		GetTempFile
		GetTempPath 			GetTempDir
		SearchPath				SearchDirs
		SetLastError			None

 */

#include "vbutil.h"
#include "win32.h"
#include "unitran.h"

// Turn off inline assembler optimization warning
#pragma warning(disable : 4704)

HINSTANCE hInst;

#define cbTempMax 256
#define cbDosPathMax 144

#if !defined(WIN32)


/* FlushFileBuffers - Win32 emulation routine to flush buffers of an 
 * open file handle to disk.
 *
 * Input:  hFile - Handle to be flushed
 *
 * Return: TRUE or FALSE
 *
 * Errors: Any of the following:
 *				ERROR_INVALID_HANDLE = 6
 */

UINT DLLAPI FlushFileBuffers(
	UINT hFile
    )
{	       
	UINT f = !_commit(hFile);
	if (f) {
		SetLastError(ERROR_INVALID_HANDLE);
	}
	return f;				
}     

/* GetDiskFreeSpace - Win32 emulation routine to get information about
 * the size of a specified disk drive.
 *
 * Input:  lpRootPathName - If 0, assume current drive, else root
 *		   directory in form:
 *
 *				d:\[path\]
 *				\[path\]
 *				\\server\share\[path\]
 *
 *		   lpSectorsPerCluster
 *
 *		   lpBytesPerSector
 *
 *		   lpNumberOfFreeClusters
 *
 *		   lpTotalNumberOfClusters
 *
 * Return: TRUE or FALSE
 *
 * Errors: Any of the following:
 *				ERROR_INVALID_DRIVE = 15
 *				ERROR_BAD_PATHNAME	= 161
 *				ERROR_NOT_SUPPORTED = 50
 */

BOOL DLLAPI GetDiskFreeSpace(
	LPCSTR lpRootPathName,
	LPDWORD lpSectorsPerCluster,
	LPDWORD lpBytesPerSector,
	LPDWORD lpNumberOfFreeClusters,
	LPDWORD lpTotalNumberOfClusters
	)
{
	struct _diskfree_t df;
	int err;
	int iDrive = ValidateDiskDrive(lpRootPathName);

	if (iDrive == INVALID_ROOT) {
		SetLastError(ERROR_INVALID_DRIVE);
		return (FALSE);
	}

	// This should get the size of a network drive not associated with a
	// drive letter, but I don't know a network-independent way of doing
	// this.
    if (iDrive == NETWORK_PATH) {
		SetLastError(ERROR_NOT_SUPPORTED);
		return (DRIVE_UNKNOWN);
	}

	// Get size from MS-DOS and translate to Windows format
	err = _dos_getdiskfree(iDrive, &df);
	if (err != 0) {
		SetLastError(ERROR_INVALID_DRIVE);
		return (FALSE);
	}
	*lpSectorsPerCluster =		df.sectors_per_cluster;
	*lpBytesPerSector = 		df.bytes_per_sector;
	*lpNumberOfFreeClusters =	df.avail_clusters;
	*lpTotalNumberOfClusters =	df.total_clusters;

	return ( TRUE );
}

/* GetDriveType - Win32 emulation routine to get information about
 * the type of a specified disk drive. This routine simply converts a
 * root path to a drive type number and calls 16-bit Windows GetDriveType.
 *
 * Input:  lpRootPathName - If 0, assume current drive, else root
 *		   directory in form:
 *
 *				d:\[path\]
 *				\[path\]
 *				\\server\share\[path\]
 *
 * Return: Any of the following:
 *				DRIVE_UNKNOWN	0
 *				DRIVE_NO_ROOT	1
 *				DRIVE_REMOVABLE 2
 *				DRIVE_FIXED 	3
 *				DRIVE_REMOTE	4
 *				DRIVE_CDROM 	5
 *				DRIVE_RAMDISK	6
 *
 * Errors: Any of the following:
 *				ERROR_INVALID_DRIVE = 15
 *				ERROR_BAD_PATHNAME	= 161
 *				ERROR_NOT_SUPPORTED = 50    
 *
 * Note: GetDriveType is called GetDriveTypeA in the source file to
 * distinguish it from Win16 GetDriveType. It has an alias to GetDriveType
 * in the .DEF file so that callers can see it as GetDriveType.
 */

UINT DLLAPI GetDriveTypeA(
	LPCSTR lpRootPathName
	)
{                       
	char lpWild[10];
	struct _find_t fi;
	UINT iType = 0;
	int iDrive = ValidateDiskDrive(lpRootPathName);

	if (iDrive == INVALID_ROOT) {
		SetLastError(ERROR_INVALID_DRIVE);
		return (DRIVE_NO_ROOT);
	}

	// Network drive names always remote
    if (iDrive == NETWORK_PATH) {
		return (DRIVE_REMOTE);
	}

	// Win32 recognizes RAM and CD-ROM formats ignored by Win16.

	// Recognize RAM disks by their volume names. The convention of
	// giving RAM disks one of the following names is generally followed
	// by RAM disk authors. Windows and other operating systems expect
	// these names.
	lstrcat(lstrcpy(lpWild, lpRootPathName), "*.*");
	if (_dos_findfirst(lpWild, _A_VOLID, &fi) == 0) {
		if (_fstricmp(fi.name, "MS-RAMDRIVE") == 0) {
			return (DRIVE_RAMDISK);
		}
		if (_fstricmp(fi.name, "RDV") == 0) {
			return (DRIVE_RAMDISK);
		}
		if (_fstricmp(fi.name, "VDISK") == 0) {
			return (DRIVE_RAMDISK);
		}
	}

	// Identify CD-ROM drives with multiplex interrupt.
	__asm {
		mov 	cx, iDrive			; Store drive
		dec 	cx					; Adjust from one-based to zero-based
		sub 	bx, bx				; Zero bx before call
		mov 	ax, 150Bh			; Function 15h, subfunction 0Bh
		int 	2Fh 				; Multiplex interrupt
		cmp 	bx, 0ADADh			; If not magic number
		jne 	NoCDROM 			;	no CD-ROM driver
		or		ax, ax				; If zero,
		je		NoCDROM				;	CD-ROM driver, but not this drive
		mov 	iType, DRIVE_CDROM	; Yes!
	NoCDROM:
	}
	if (iType) {
		return (iType);
	}

	// Call Windows API to check other possibilities.
	iType = GetDriveType(iDrive - 1);
	
	return iType;
	
}	

/* GetFullPathName - Win32 API emulation routine to get the fully
 * qualified path name of a specified name.
 *
 * Input:  lpFileName - file to be qualified in one of these formats:
 *
 *				[relpath\]file[.ext]
 *				\[path\]file[.ext]
 *				.\[path\]file[.ext]
 *				d:\[path\]file[.ext]
 *				..\[path\]file[.ext]
 *				\\server\machine\[path\]file[.ext]
 *
 *		   nBufferLength - length of buffer to receive the name
 *
 *		   lpBuffer - buffer to receive the name
 *
 *		   lpFilePart - address of the file name component of the
 *		   qualified name
 *
 * Return: 0 if error or the length of the buffer (excluding the null).
 *		   If the length of the qualified name would be greater than
 *		   the buffer length, the required length is returned.
 *
 * Errors: Any of the following:
 *				ERROR_BUFFER_OVERFLOW	   = 111
 *				ERROR_INVALID_DRIVE 	   = 15
 *				ERROR_CALL_NOT_IMPLEMENTED = 120
 *				ERROR_BAD_PATHNAME		   = 161
 */

DWORD DLLAPI GetFullPathName(
    LPCSTR lpFileName,
	DWORD cbBuffer,
    LPSTR lpBuffer,
    LPSTR *lpFilePart
	)
{
	DWORD cb;
	LPSTR lp;
	LPSTR lpTemp;
	HGLOBAL hTemp; 
	
	if ((lpFileName == NULL) || (lpBuffer == NULL)) {
		SetLastError(ERROR_BAD_PATHNAME);	
	 	return (0);
	}

	// Memory must be allocated from Windows
	if ((hTemp = GlobalAlloc(GPTR, cbTempMax)) == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
	 	return (0);
	}
	lpTemp = GlobalLock(hTemp);
	if (_fullpath(lpTemp, lpFileName, cbTempMax ) == NULL) {
		GlobalUnlock(hTemp);
		GlobalFree(hTemp);
		return (0);
	}
	cb = lstrlen(lpTemp);
	if (cb <= cbBuffer) {
		lstrcpy(lpBuffer, lpTemp);
		lp = lpBuffer + cb;
		while ((lp >= lpBuffer) && (*lp != '\\') && (*lp != ':')) {
			lp--;
		}	
		*lpFilePart = lp + 1;  
	} else {
		// Allow one extra for null
		cb++;
	}
	
	GlobalUnlock(hTemp);
	GlobalFree(hTemp);
   	return ( cb );
}

/* GetLastError - Win32 API emulation routine to get the number of the
 * last error received by a function.
 *
 * Input:  None
 *
 * Return: A non-zero error code (zero indicates no error has occurred)
 *
 * Errors: None possible
 */

DWORD DLLAPI GetLastError(
    VOID
	)
{
   return ( dwLastError );
}

/* GetLogicalDrives - Win32 emulation routine to get a map of all
 * available drives.
 *
 * Input:  None
 *                                                     
 * Return: Bit map with bits 0-25 representing presence of each drive.
 */
DWORD DLLAPI GetLogicalDrives(
    VOID
	)
{
	// This could be simpler using GetDriveType, but it looks like this was
	// defined to be unable to tell the difference between an unknown drive
	// and a non-existent drive.
	DWORD dwDisks = 0;
	WORD iCur, i, f;

	__asm {
        mov     ah, 19h                 ; Get current disk
		int		21h
		sub		ah, ah					; Clear top
		mov 	iCur, ax				; Save initial disk
	}

	for (i = 0; i < 26; i++) {

		__asm {
			sub 	cx, cx					; Assume disk doesn't exist
			mov 	dx, i					; Set drive number
			mov 	ah, 0Eh 				; Select new drive
			int		21h
			mov 	ah, 19h 				; Get current drive
			int		21h
			cmp		al, dl					; See if the same
			jne 	skip1					; Skip if drive is invalid
			inc 	cx						; Set disk exist flag
		skip1:
			mov 	f, cx					; Store flag
		}
		dwDisks |= ((DWORD)f << i);
	}

	__asm {
		mov 	dx, iCur					; Restore original disk (to DL)
		mov		ah, 0Eh
		int		21h
	}

	return (dwDisks);
}


/* GetLogicalDrivesStrings - Win32 emulation routine to get an array
 * of strings containing all available drives in the format "d:\". 
 *
 * Input: 	cchBuffer	- Specifies the maximum size, in characters, 
 *			of buffer pointed to by lpszBuffer. This size does not 
 *			include the terminating null character. 
 *
 *        	lpszBuffer - Points to a buffer that receives a series of 
 * 		  	null-terminated strings, one for each valid drive in the 
 * 		  	system, that end with a second null character. The 
 *		  	following example shows the buffer contents with <null> 
 * 		  	representing the null terminator. 
 *
 *				c:\<null>d:\<null><null>
 *                                                     
 * Return: 	If the function succeeds, the return value is the length, 
 *		   	in characters, of the strings copied to the buffer, not 
 *			including the terminating null character. 
 *
 *			If the buffer is not large enough, the return value is 
 *			greater than cchBuffer. It is the size of the buffer 
 *			required to hold the drive strings. If the function 
 *			fails, the return value is zero. To get extended error 
 *			information, use the GetLastError function. 
 */

DWORD DLLAPI GetLogicalDriveStrings(
	DWORD cchBuffer,
    LPSTR lpszBuffer
	)
{
	int i;
	DWORD cch = 0;
	char szDrive[4] = "?:\\";
	// Get the drives in the system
	DWORD dwDisks = GetLogicalDrives();

	// See if they'll fit
	for (i = 0; i < 26; i++) {
		if ((dwDisks << i) & 1) {
			cch += 4;
		}
	}
	// One extra for final null
	cch++;

	// If not, return the size without touching the buffer.
	if (cch > cchBuffer) {
		return (cch);
	}

	// Stuff 'em in
	cch = 0;
	for (i = 0; i < 26; i++) {
		if ((dwDisks << i) & 1) {
			cch += 4;
			szDrive[0] = 'A' + i;
			strcat(lpszBuffer, szDrive);
		}
	}
	// Add second null
	lpszBuffer[cch] = '\0';

	// Return length
	return (cch);
}


/*
 * Note: GetTempFileName is called GetTempFileNameA in the source file
 * to distinguish it from Win16 GetTempFileName. It has an alias to
 * GetTempFileName in the .DEF file so that callers can see it as
 * GetTempFileName.
 */
UINT DLLAPI GetTempFileNameA(
    LPCSTR lpPathName,
    LPCSTR lpPrefixString,
    UINT uUnique,
    LPSTR lpTempFileName
	)
{                            
	#define lpPathCur (lpPath + cbDosPathMax - 1)
	char lpPath[cbDosPathMax * 2];
	BYTE iDrive;
	_fullpath(lpPath, lpPathName, cbDosPathMax);
	iDrive = toupper(lpPath[0]) - 'A' + 1;
	if (iDrive > 26 || iDrive < 1) {
		return (0);
	}
	_getcwd(lpPathCur, cbDosPathMax);
	_chdrive(iDrive);
	_chdir(lpPath);

	uUnique = GetTempFileName(TF_FORCEDRIVE, lpPrefixString, uUnique, lpPath);
	_fullpath(lpTempFileName, lpPath, cbDosPathMax);
	iDrive = toupper(lpPathCur[0]) - 'A' + 1;
	_chdrive(iDrive);
	_chdir(lpPathCur);
	return (uUnique);
}

/* GetTempPath - Win32 API emulation routine to get the directory for
 * temporary files. This directory is:
 *
 *	  1. Directory of the TEMP environment variable
 *	  2. Directory of the TMP environment variable
 *	  3. Current directory
 *
 * Input:  lpBuffer - Pointer to buffer for directory string received
 *		   in one of the following formats:
 *
 *				d:\[path]
 *				\\server\share\[path]
 *
 *		   nBufferLength - Length of buffer for directory string
 *
 * Return: Length of the buffer path string. If the path name won't
 *		   fit, it is the length required for the path name.
 *
 * Errors: Any of the following:
 *				ERROR_BUFFER_OVERFLOW = 111
 */

DWORD DLLAPI GetTempPath(
	DWORD cbBuffer,
    LPSTR lpBuffer
	)
{
	DWORD cb;
	LPSTR pch;
	LPSTR lpTemp;
	HGLOBAL hTemp;
	
	// Return the TEMP variable if it exists
	cb = GetEnvironmentVariable("TEMP", lpBuffer, cbBuffer);
	if (cb != 0) {
		cb = NormalizeDir(lpBuffer, cbBuffer);
		return (cb);
	}
	
	// Return the TMP variable if it exists
	cb = GetEnvironmentVariable("TMP", lpBuffer, cbBuffer);
	if (cb != 0) {
		cb = NormalizeDir(lpBuffer, cbBuffer);
		return (cb);
	}
	
	if ((hTemp = GlobalAlloc(GPTR, cbTempMax)) == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
	 	return (0);
	}
	lpTemp = GlobalLock(hTemp);
	// Return the current directory
	pch = getcwd(lpTemp, cbTempMax - 1);
	ASSERT(pch != NULL);
	cb = lstrlen(lpTemp);
	// Append \ if necessary
	if (lpTemp[cb] != '\\') {
		lstrcat(lpTemp, "\\");
	}
	
	// Copy the string if there's space, otherwise set error
	if (cb <= cbBuffer) {
		lstrcpy(lpBuffer, lpTemp);
	}	
	else {
		SetLastError(ERROR_BUFFER_OVERFLOW);
	}
	GlobalUnlock(hTemp);				
	GlobalFree(hTemp);				
	return (cb);
}

/* SearchPath - Win32 API emulation routine to search for a specified file
 * starting from a specified path. It returns a fully-qualified path name
 * for the found file.
 *
 * Input:  lpPath - The search path to be used. This is a string
 *	   containing paths separated by semicolons:
 *
 *			   path;path;path[;]
 *
 *		   If path is null, the default path is:
 *
 *			   1. The executable directory
 *			   2. The current directory
 *			   3. The windows directory
 *			   4. The windows system directory
 *			   5. The directory listed in the PATH environment variable
 *
 *		   Note that if a path is provided, the current directory will
 *		   not be searched (unless it is in the path).
 *
 *		   lpFileName - The file name to search for. If absolute path,
 *		   lpPath and lpExtension are ignored.
 *
 *		   lpExtension - An optional extension to be added if the file
 *		   name does not have an extension.
 *
 *		   nBufferLength - Length of buffer to receive spec of found file.
 *
 *		   lpBuffer - Buffer to receive the fully-qualified spec of the
 *		   found file.
 *
 *		   lpFilePart - Pointer to the name component of the found file.
 *
 * Return: Length of the buffer path string. If the path name won't
 *		   fit, it is the length required for the path name. If the
 *		   no file is found, 0 is returned.
 *
 * Errors: Any of the errors passed by SearchPath including:
 *				ERROR_BUFFER_OVERFLOW	   = 111
 */

DWORD DLLAPI SearchPath(
    LPCSTR lpPath,
    LPCSTR lpFileName,
    LPCSTR lpExtension,
	DWORD cbBuffer,
    LPSTR lpBuffer,
    LPSTR *lpFilePart
	)
{
	char lpWork[cbDosPathMax];
	unsigned cb, i;
	BOOL f;
	HTASK hTask;   
	TASKENTRY te; 
	LPSTR lpFileNameExt;
	LPSTR lp;
	
	if ((*lpFileName == '\0') || (lpFileName == NULL)) {
		return (NULL);
	}
	
	// Append extension if needed 
	if (lpExtension != NULL) {
		for (lp = (LPSTR)lpFileName + lstrlen(lpFileName) - 1;;lp--) {
			// If . encountered first, it already has extension
			if (*lp == '.') {
				break;
			}       
			// If \ or : or start of string encountered first, it has no extension
			if ((*lp == '\\') || (*lp == ':') || (lp == lpFileName)) {
				lp = (LPSTR)lpFileName;
				break;					
			}	
		}
		// Append extension and set lpFileName to new copy
		if (lp == lpFileName) {
			lpFileNameExt = (LPSTR)alloca(lstrlen(lpFileName) +
										  lstrlen(lpExtension) + 3);
			if (lpFileNameExt == NULL) {
				SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				return (0);
			}

			lstrcpy(lpFileNameExt, lpFileName);
			if (*lpExtension != '.') {
				lstrcat(lpFileNameExt, ".");
			}
			lstrcat(lpFileNameExt, lpExtension);
			lpFileName = lpFileNameExt;
		}		
	}	
	
	if (lpPath == NULL) {
    
		// Check for already full path
		if (IsFullPath(lpFileName)) {
			if (!access(lpFileName, 0)) {
				cb = lstrlen(lpFileName);
				if (cb <= cbBuffer) {
					lstrcpy(lpBuffer, lpFileName);
					lp = lpBuffer + cb - 1;
					for (;*lp != '\\';lp--) {
						;
					}
					*lpFilePart = lp + 1;
				}
				return cb;
			} else {
				return (0);
			}
		}

		// Check in current directory 
		getcwd(lpWork, cbDosPathMax);
		if (cb = ExistFilePath(lpWork, lpFileName, lpBuffer, cbBuffer, lpFilePart)) {
			return (cb);
		}

		// Check in Windows directory
		cb = GetWindowsDirectory(lpWork, cbDosPathMax);
	 	if (cb = ExistFilePath(lpWork, lpFileName, lpBuffer, cbBuffer, lpFilePart)) {
			return (cb);
		}

		// Check in Systems directory
		cb = GetSystemDirectory(lpWork, cbDosPathMax);
		if (cb = ExistFilePath(lpWork, lpFileName, lpBuffer, cbBuffer, lpFilePart)) {
			return (cb);
		}
		
		// Check in the executable load directory
    	hTask = GetCurrentTask();  
    	ASSERT(hTask != NULL);
    	te.dwSize = sizeof(TASKENTRY);
    	f = TaskFindHandle(&te, hTask);
    	ASSERT(f);
		cb = GetModuleFileName(te.hModule, lpWork, cbDosPathMax);
    	ASSERT(cb);
    	for (i = cb; (i > 0) && (lpWork[i] != '\\'); i--)
    		;          
		if (i != 0) {
			lpWork[i + 1] = '\0';
			if (cb = ExistFilePath(lpWork, lpFileName, lpBuffer, cbBuffer, lpFilePart)) {
				return (cb);
			}	
		}

		// Search the path
		_searchenv(lpFileName, "PATH", lpWork);
		if (cb = lstrlen(lpWork)) {
			if (cb <= cbBuffer) {
				lstrcpy(lpBuffer, lpWork);
				for (i = cb; (i > 0) && (lpWork[i] != '\\'); i--)
					;
				ASSERT (i != 0);
				*lpFilePart = lpBuffer + i + 1;
			}
			return (cb);
		}
	} else {
		LPSTR lp;
		LPSTR lpLast;
		lp = lpLast = (LPSTR)lpPath; 
		while (1) {
			if ((*lp == '\0') || (*lp == ';')) {
				_fstrncpy(lpWork, lpLast, lp - lpLast);
				lpWork[lp - lpLast] = '\0';
				lpLast = lp + 1;
				if (cb = ExistFilePath(lpWork, lpFileName,
									   lpBuffer, cbBuffer, lpFilePart)) {
					return (cb);
				}
			}
			if (*lp == '\0') {
				break;
			}	
			lp++;
		} 
	}

    return ( 0 );
}

/* SetLastError - Win32 API emulation routine to set the number to be
 * received by GetLastError.
 *
 * Each thread stores just one error code and error string. The
 * code and string stored for a thread represent the most recent
 * error.
 *
 * Windows32 API functions that return indications of failure call
 * SetLastError when they fail. They do not call it when they
 * succeed.  Thus, if three Win32 API function calls are made, and
 * the first one fails and the second two succeed, the error code and
 * string stored by the first one are still available after the second
 * two succeed.
 *
 * SetLastError is intended primarily for DLLs, not application code.
 *
 * Input:  dwError - Error code to set (usually NO_ERROR)
 *
 * Return: none
 */

VOID DLLAPI SetLastError(
	DWORD dwError
	)
{
	dwLastError = dwError;
	return;
}

// Internal routines used by Win16 only routines

/* GetEnvironmentVariable - Win32 emulation routine to get the string
 * of a specified environment variable. This is for internal use only in 
 * C since Basic already has Environ$ function. 
 *
 * Input:  lpName - String name to be found in the environment table
 *	   lpBuffer - Buffer to receive the environment string
 *	   nSize - Maximum size of the buffer
 *
 *
 * Return: The actual number of bytes stored in the memory pointed to
 *	   by the lpBuffer parameter. The return value is zero if the
 *	   environment variable name was not found in the current
 *	   process's environment. The returned value does not include
 *	   the null terminator byte.
 *
 * Errors: Including the following:
 *		ERROR_BUFFER_OVERFLOW	= 111
 *		ERROR_ENVVAR_NOT_FOUND	= 203
 */

DWORD DLLAPI GetEnvironmentVariable(
	LPCSTR lpName,
	LPSTR lpBuffer,
	DWORD cbSize
	)
{
  	LPSTR lpValue;
	DWORD cbValue = 0;
	
	lpValue = GetEnv(lpName);
	
	if (lpValue == NULL) {
		if (cbSize) {
			*lpBuffer = '\0';
		}
		return (0);
	}	
	
	cbValue = lstrlen(lpValue) + 1;
	
	if (cbValue <= cbSize) {
		lstrcpy(lpBuffer, lpValue);
		cbValue--;
	}	
	return (cbValue);
}


LPSTR DLLAPI GetEnv(
	LPCSTR lpName
	)
{
	LPSTR lp;
	LPSTR lpVar;
	if ((lpName == NULL) || (*lpName == '\0')) {
		return (NULL);
	}

	//	Get a pointer to the MS-DOS environment block
	lp = GetDOSEnvironment ();
	while (*lp)	//	While there are strings to parse
	{
		// Make a copy of the pointer to the name of the
		// environment variable to search for.
		lpVar = (LPSTR)lpName;

		//	Check to see if the variable names match
		while (*lp && *lpVar && *lp == *lpVar) {
			lp++;
			lpVar++;
		}

		if ((*lp == '=') && (*lpVar == '\0')) {
			// If the names match, the environment pointer is on the "="
			// and the name pointer is on null terminator. Increment to the
			// contents and return.

			return (lp + 1);
		} else {
			// Move to end of string
			while (*lp ) {
				lp++;
			}
		}
		// Move to next string
		lp++;
	}
	// Variable not found
	return (NULL);
}

/* NormalizeDir - Makes sure that a directory ends with a backslash. If not
 * enough room, returns the length that would be required. 
 *
 */

DWORD DLLAPI NormalizeDir(
    LPSTR lpBuffer,
	DWORD cbBuffer
	)
{
	DWORD cb = strlen(lpBuffer);
	if (lpBuffer[cb - 1] != '\\') {
		cb++;
		if (cb < cbBuffer) {
			lpBuffer[cb - 1] = '\\';
		}	
	}
	return (cb);
}

/* ValidateDiskDrive - Validates a root path name, indicating that the name
 * is valid (but not that the drive exists). Returns one of the following:
 *
 *		0				The current drive
 *		1-26			Drive A to Z
 *		INVALID_ROOT	Invalid root path
 *		NETWORK_PATH	Network drive name
 */

int ValidateDiskDrive(
	LPCSTR lpRootPathName
	)
{
	int i;
	char const *pch;

	// Null indicates current drive (drive 0)
	if (lpRootPathName == NULL) {
		return (0);
	}
	i = toupper(lpRootPathName[0]);

	// Check for drive root in the format "d:\" and return A=1, B=2, etc.
	if ((lpRootPathName[1] == ':') &&
		(lpRootPathName[2] == '\\') &&
		(i >= 'A') &&
		(i <= 'Z') &&
		(lpRootPathName[3] == '\0')) {
		return (i - 'A' + 1);
	}

	// Check for valid network name in format "\\server\share\"
	if ((lpRootPathName[0] != '\\') ||		// First two characters "\\"
		(lpRootPathName[1] != '\\') ||
		(lpRootPathName[2] == '\\')) {		// Next character not "\"
		return (INVALID_ROOT);
	}

	// Count the backslashes
	for (i = 0, pch = lpRootPathName + 2; *pch != '\0'; pch++) {
		if (*pch == '\\') {
			i++;
		}
	}
	// There must be two backslashes including the last character
	if ((i != 2) || (*(pch - 1) != '\\')) {
		return (INVALID_ROOT);
	}
	return (NETWORK_PATH);
}

/* IsFullPath - Validates a root path name, indicating that the name
 * is valid (but not that the drive exists). Returns one of the following:
 *
 *		0				The current drive
 *		1-26			Drive A to Z
 *		INVALID_ROOT	Invalid root path
 *		NETWORK_PATH	Network drive name
 */

BOOL IsFullPath(
	LPCSTR lpPathName
	)
{
	char ch;

	ch = lpPathName[0];
	ch = toupper(ch);

	// Check for full path in drive format "d:\..."
	if ((lpPathName[1] == ':') &&
		(lpPathName[2] == '\\') &&
		(ch >= 'A') &&
		(ch <= 'Z')) {
		return (TRUE);
	}

	// Check for full path in network format "\\..."
	if ((lpPathName[0] == '\\') ||		// First two characters "\\"
		(lpPathName[1] == '\\')) {
		return (TRUE);
	}
    
    return (FALSE);
}
            
// Combine a path and a filename and see if the file exists. If so, return
// its length and a pointer to its base portion. 
unsigned ExistFilePath(
	LPSTR lpPath,
	LPCSTR lpFileName,
	LPSTR lpBuffer,
	DWORD cbBuffer, 
	LPSTR * lpFilePart)
{
	unsigned cb, iFile;
	cb = lstrlen(lpPath);
	if (lpPath[cb - 1] != '\\') {
		lpPath[cb++] = '\\';
		lpPath[cb] = '\0';
	}
	lstrcpy(lpPath + cb, lpFileName);

	if (!access(lpPath, 0)) {
		iFile = cb;
		cb = lstrlen(lpPath);
		if (cb <= cbBuffer) {
			lstrcpy(lpBuffer, lpPath);
			*lpFilePart = lpBuffer + iFile;
		}
		return cb;
	} else {
		return (0);
	}
}
		
#endif // !defined(WIN32)


// Win32 functions with Basic interface

/* GetFullPath - Basic version of Win32 API emulation routine. It returns a
 * BSTR, and indexes to the file name, directory, and extension parts of the 
 * full name. 
 *
 * Input:  sFileName - file to be qualified in one of these formats:
 *
 *				[relpath\]file.ext
 *				\[path\]file.ext
 *				.\[path\]file.ext
 *				d:\[path\]file.ext
 *				..\[path\]file.ext
 *				\\server\machine\[path\]file.ext
 *			iName - variable to receive file name position
 *			iDir - variable to receive directory position
 *			iExt - variable to receive extension position
 *
 * Return: Full path name, or an empty string on failure
 *
 * Errors: Any of the following:
 *				ERROR_BUFFER_OVERFLOW	   = 111
 *				ERROR_INVALID_DRIVE 	   = 15
 *				ERROR_CALL_NOT_IMPLEMENTED = 120
 *				ERROR_BAD_PATHNAME		   = 161
 */

BSTR DLLAPI GetFullPath(
    BSTR FAR * psFileName,
	LPDWORD lpdwDirPart,
	LPDWORD lpdwFilePart,
	LPDWORD lpdwExtPart
	)
{
	LPSTR lpFilePart;
	BSTR sFileNameA = GetAFromU(*psFileName);
	DWORD cb;
	char ach[2];
	BSTR sRetA;
	BSTR sRetU;

	// Get the path size, then allocate space for it
	cb = GetFullPathName((LPSTR)sFileNameA, 1, ach, &lpFilePart);
	sRetA = SysAllocStringByteLen(NULL, (UINT)cb - 1);
	if (sRetA == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
		FreeUA(sFileNameA);
	 	return (NULL);
	}
	cb = GetFullPathName((LPSTR)sFileNameA, cb, (LPSTR)sRetA, &lpFilePart);
	FreeUA(sFileNameA);
	ASSERT(cb <= cbTempMax);
	*lpdwFilePart = (DWORD)0;
	*lpdwDirPart = (DWORD)0;
	*lpdwExtPart = (DWORD)0;
	if ((sRetA != NULL) && cb) {
		*lpdwFilePart = lpFilePart - (LPSTR)sRetA + 1;
		GetDirExt((LPSTR)sRetA, lpdwDirPart, lpdwExtPart);
	}
	sRetU = GetUFromA(sRetA);
	FreeUA(sRetA);
	return (sRetU);
}

BSTR DLLAPI GetTempFile(
    BSTR FAR * psPathName,
	BSTR FAR * psPrefixString
	)
{
	DWORD cb;
	BSTR sPathNameA = GetAFromU(*psPathName);
	BSTR sPrefixStringA = GetAFromU(*psPrefixString);
	BSTR sTmpA;
	BSTR sRetA;
	BSTR sRetU;

	sTmpA = SysAllocStringByteLen(NULL, (UINT)cbTempMax);
	if (sTmpA == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
		FreeUA(sPathNameA);
		FreeUA(sPrefixStringA);
	 	return (NULL);
	}
	cb = GetTempFileNameA((LPSTR)sPathNameA, (LPSTR)sPrefixStringA, 
						  0, (LPSTR)sTmpA);
	cb = lstrlen((LPSTR)sTmpA);
	sRetA = SysAllocStringByteLen((LPSTR)sTmpA, (UINT)cb);
	FreeUA(sTmpA);
	FreeUA(sPathNameA);
	FreeUA(sPrefixStringA);
	if (sRetA == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
	 	return (NULL);
	}
	sRetU = GetUFromA(sRetA);
	FreeUA(sRetA);
	return (sRetU);
}

BSTR DLLAPI GetTempDir(
	void
	)
{
	DWORD cb;
	BSTR sRetA;
	BSTR sRetU;
	char ach[2];
	LPSTR lpTemp = ach;

	cb = GetTempPath(1, ach);
	sRetA = SysAllocStringByteLen(NULL, (UINT)cb - 1);
	if (sRetA == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
	 	return (0);
	}
	cb = GetTempPath(cb, (LPSTR)sRetA);
	sRetU = GetUFromA(sRetA);
	FreeUA(sRetA);
	return (sRetU);
}

BSTR DLLAPI SearchDirs(
    BSTR FAR * psPath,
    BSTR FAR * psFileName,
    BSTR FAR * psExtension,
	LPDWORD lpdwDirPart,
	LPDWORD lpdwFilePart,
	LPDWORD lpdwExtPart
	)
{
	LPSTR lpFilePart;
	DWORD cb;
	BSTR sPathA = GetAFromU(*psPath);
	BSTR sFileNameA = GetAFromU(*psFileName);
	BSTR sExtensionA = GetAFromU(*psExtension);
	BSTR sTmpA;
	BSTR sRetA;
	BSTR sRetU;
	
  	if ((sFileNameA == NULL) || (*sFileNameA == '\0')) {
		FreeUA(sPathA);
		FreeUA(sFileNameA);
		FreeUA(sExtensionA);
		return (NULL);
	}
	
	// Treat empty strings the same as null pointers
	if ((sPathA != NULL) && (*sPathA == '\0')) {
		FreeUA(sPathA);
		sPathA = NULL;
	}
	if ((sExtensionA != NULL) && (*sExtensionA == '\0')) {
		FreeUA(sExtensionA);
		sExtensionA = NULL;
	}
	
	sTmpA = SysAllocStringByteLen(NULL, (UINT)cbTempMax);
	if (sTmpA == NULL) {
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);	
		FreeUA(sPathA);
		FreeUA(sFileNameA);
		FreeUA(sExtensionA);
	 	return (NULL);
	}

	// Get the file
	cb = (DWORD)SearchPath((LPSTR)sPathA, (LPSTR)sFileNameA, 
						   (LPSTR)sExtensionA, (DWORD)cbTempMax, 
						   (LPSTR)sTmpA, &lpFilePart);
	ASSERT(cb <= cbTempMax);
	FreeUA(sPathA);
	FreeUA(sFileNameA);
	FreeUA(sExtensionA);
	sRetA = SysAllocStringByteLen((LPSTR)sTmpA, (UINT)cb);

	// Calculate the file part offsets
	*lpdwFilePart = (DWORD)0;
	*lpdwDirPart = (DWORD)0;
	*lpdwExtPart = (DWORD)0;
	if ((sRetA != NULL) && cb) {
		*lpdwFilePart = lpFilePart - (LPSTR)sTmpA + 1;
		GetDirExt((LPSTR)sRetA, lpdwDirPart, lpdwExtPart);
	} 
	FreeUA(sTmpA);
	sRetU = GetUFromA(sRetA);
	FreeUA(sRetA);
	return (sRetU);
}	
 
void GetDirExt(
	LPSTR lpSpec, 
	LPDWORD lpdwDirPart,
	LPDWORD lpdwExtPart
	)
{	
	LPSTR lp;
	char ch = toupper(*lpSpec);
	
	// If in format d:\path\name.ext, return 3
	if ((ch >= 'A') && 
		(ch <= 'Z') &&
		(*(lpSpec+1) == ':') &&
		(*(lpSpec+2) == '\\')) {
		*lpdwDirPart = 3;
	// If in format \\machine\share\path\name.ext, return position of \path 		
	} else {
		ASSERT(ch == '\\');
		ASSERT(*(lpSpec+1) == '\\');
		for (lp = lpSpec + 2;*lp && (*lp != '\\');lp++)
			;
		for (;*lp && (*lp != '\\');lp++)
			;
		*lpdwDirPart = lp - lpSpec;
	}
	// Find extension
	*lpdwExtPart = lstrlen(lpSpec); // Assume no extension 
	for (lp = lpSpec + lstrlen(lpSpec);lp > lpSpec; lp--) {
		if (*lp == '.') {
			*lpdwExtPart = lp - lpSpec + 1;
			break;
		}
		if ((*lp == '\\') || (*lp == ':')) {
			break;
		}
	}
}					
	
	
